/*
 *  Copyright (C) 2008 dhewg, #wiidev efnet
 *
 *  this file is part of wiifuse
 *  http://wiibrew.org/index.php?title=Wiifuse
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gccore.h>
#include <ogcsys.h>

#include "../../shared.h"
#include "global.h"
#include "network.h"
#include "server.h"
#include <wiiuse/wpad.h>

static void *xfb = NULL;
static GXRModeObj *rmode;

// TODO reset/power cb
void (*reload)() = (void (*) ()) 0x80001800;

int main() {
        WPAD_Init ();
        VIDEO_Init ();

        switch (VIDEO_GetCurrentTvMode())
        {
                case VI_NTSC:
                        rmode = &TVNtsc480IntDf;
                        break;
                case VI_PAL:
                        rmode = &TVPal528IntDf;
                        break;
                case VI_MPAL:
                        rmode = &TVMpal480IntDf;
                        break;
                default:
                        rmode = &TVNtsc480IntDf;
                        break;
        }

        xfb = MEM_K0_TO_K1 (SYS_AllocateFramebuffer (rmode));

        VIDEO_Configure (rmode);
        VIDEO_SetNextFramebuffer (xfb);
        VIDEO_SetBlack (FALSE);
        VIDEO_Flush ();
        VIDEO_WaitVSync ();
        if (rmode->viTVMode&VI_NON_INTERLACE)
                VIDEO_WaitVSync();
        console_init(xfb, 20, 64, rmode->fbWidth, rmode->xfbHeight,
                     rmode->fbWidth * 2);
        VIDEO_SetNextFramebuffer (xfb);

        printf ("wiifuse server " WIIFUSE_VERSION "\n"
                "coded by dhewg, #wiidev efnet/blitzed\n\n"
                "network by jepler\n\n"
                "running under ios version %d v%d.%d\n\n", IOS_GetVersion (),
                IOS_GetRevisionMajor (), IOS_GetRevisionMinor ());
        
        
        printf ("preparing network - press HOME to exit or wait\n"); fflush(stdout);
        while(1) {
            int result = net_init ();
            printf("result=%12d\r", result); fflush(stdout);
            if(result == 0) break;
            WPAD_ScanPads();
            if(WPAD_ButtonsDown(0) & WPAD_BUTTON_HOME) goto quit_now;
        }

        int socket = net_socket(AF_INET, SOCK_STREAM, 0);
        if(socket < 0) {
                printf("socket() -> %d\n", socket);
                goto quit;
        }

        struct sockaddr_in sa;
        memset(&sa, 0, sizeof(sa));
        sa.sin_family = AF_INET;
        sa.sin_port = htons(4875);
        sa.sin_addr.s_addr = INADDR_ANY;

        int result = net_bind(socket, (struct sockaddr*)&sa, sizeof(sa));
        if(result < 0) {
                printf("socket() -> %d\n", result);
                goto quit;
        }

        printf ("listening - press HOME to exit or start wiifuse\n"); fflush(stdout);
        do {
            result = net_listen(socket, 1);
            printf("result=%12d\r", result); fflush(stdout);
            WPAD_ScanPads();
            if(WPAD_ButtonsDown(0) & WPAD_BUTTON_HOME) goto quit_now;
            VIDEO_WaitVSync();
        } while(result < 0);

        printf("listen() -> %d\n", result);
        if(result < 0) {
                goto quit;
        }

        u32 len = sizeof(sa);
        int client = net_accept(socket, (struct sockaddr*)&sa, &len);

        if(client < 0) {
                printf("accept() -> %d\n", result);
                goto quit;
        }
        server_run (client);
        net_close(client);
        goto quit_now;

quit:
        printf("press HOME to quit\n"); fflush(stdout);

        while(1) {
            WPAD_ScanPads();
            if(WPAD_ButtonsDown(0) & WPAD_BUTTON_HOME) goto quit_now;
            VIDEO_WaitVSync();
        }

quit_now:
        printf ("reloading\n"); fflush(stdout);
        reload ();

        return 0;
}

